from pathlib import Path
from typing import cast

from pdm.core import Core
from pdm.exceptions import PdmUsageError, ProjectError
from pdm.project import Project
from tomlkit.items import Array

from pdm_conda.installers.manager import CondaInstallManager
from pdm_conda.installers.synchronizers import CondaSynchronizer
from pdm_conda.models.repositories import (
    LockedCondaRepository,
    LockedRepository,
    PyPICondaRepository,
)
from pdm_conda.models.requirements import (
    CondaRequirement,
    NamedRequirement,
    Requirement,
    parse_requirement,
)


class CondaProject(Project):
    def __init__(
        self,
        core: Core,
        root_path: str | Path | None,
        is_global: bool = False,
        global_config: str | Path | None = None,
    ) -> None:
        super().__init__(core, root_path, is_global, global_config)
        self.core.repository_class = PyPICondaRepository
        self.core.install_manager_class = CondaInstallManager
        self.locked_repository_class = LockedCondaRepository
        self.core.synchronizer_class = CondaSynchronizer
        self.virtual_packages: set[str] = set()

    def get_dependencies(self, group: str | None = None) -> dict[str, Requirement]:
        result = super().get_dependencies(group)

        settings = self.pyproject.settings.get("conda", {})
        group = group or "default"
        optional_dependencies = settings.get("optional-dependencies", {})
        dev_dependencies = settings.get("dev-dependencies", {})
        deps = []

        if group == "default":
            deps = settings.get("dependencies", [])
        else:
            if group in optional_dependencies and group in dev_dependencies:
                self.core.ui.echo(
                    f"The {group} group exists in both [optional-dependencies] "
                    "and [dev-dependencies], the former is taken.",
                    err=True,
                    style="warning",
                )
            if group in optional_dependencies:
                deps = optional_dependencies[group]
            elif group in dev_dependencies:
                deps = dev_dependencies[group]
            elif not result:
                raise PdmUsageError(f"Non-exist group {group}")

        for line in deps:
            req = parse_requirement(f"conda:{line}")
            # search for package with extras to remove it
            pypi_req = next((v for v in result.values() if v.name == req.name), None)
            if pypi_req is not None:
                result.pop(pypi_req.identify())
                if not req.specifier:
                    req.specifier = pypi_req.specifier
            result[req.identify()] = req

        if self.pyproject.settings.get("conda", {}).get("as_default_manager", False):
            for k in list(result):
                req = result[k]
                if isinstance(req, NamedRequirement):
                    result.pop(k)
                    req.extras = None
                    result[k] = parse_requirement(f"conda:{req.as_line()}")
        return result

    @property
    def locked_repository(self) -> LockedRepository:
        try:
            lockfile = self.lockfile._data.unwrap()
        except ProjectError:
            lockfile = {}

        return self.locked_repository_class(lockfile, self.sources, self.environment)

    def get_conda_pyproject_dependencies(self, group: str, dev: bool = False) -> list[str]:
        """
        Get the conda dependencies array in the pyproject.toml
        """
        settings = self.pyproject.settings.setdefault("conda", dict())
        if group == "default":
            return settings.setdefault("dependencies", [])
        name = "optional" if not dev else "dev"
        return settings.setdefault(f"{name}-dependencies", dict()).setdefault(group, [])

    def add_dependencies(
        self,
        requirements: dict[str, Requirement],
        to_group: str = "default",
        dev: bool = False,
        show_message: bool = True,
    ) -> None:
        conda_requirements = {n: r for n, r in requirements.items() if isinstance(r, CondaRequirement)}
        if conda_requirements:
            deps = self.get_conda_pyproject_dependencies(to_group, dev)
            cast(Array, deps).multiline(True)
            for _, dep in conda_requirements.items():
                matched_index = next((i for i, r in enumerate(deps) if dep.matches(f"conda:{r}")), None)
                req = dep.as_line(with_channel=True)
                if matched_index is None:
                    deps.append(req)
                else:
                    deps[matched_index] = req

        requirements = {n: r for n, r in requirements.items() if n not in conda_requirements} | {
            n: r.as_named_requirement() for n, r in conda_requirements.items() if r.is_python_package
        }
        super().add_dependencies(requirements, to_group, dev, show_message)



import dataclasses
import functools
import re
from typing import Any

from pdm._types import RequirementDict
from pdm.models.markers import get_marker
from pdm.models.requirements import NamedRequirement, PackageRequirement, Requirement, T
from pdm.models.requirements import parse_requirement as _parse_requirement

_conda_req_re = re.compile(r"conda:((\w+::)?.+)$")
_specifier_re = re.compile(r"[<>=~!].*")

_patched = False


@dataclasses.dataclass(eq=False)
class CondaRequirement(NamedRequirement):
    channel: str | None = None

    @property
    def project_name(self) -> str | None:
        return self.name

    @property
    def is_python_package(self) -> bool:
        return not self.requires_python

    @classmethod
    def create(cls: type[T], **kwargs: Any) -> T:
        kwargs.pop("conda_managed", None)
        return super().create(**kwargs)

    def as_line(self, with_channel=False) -> str:
        channel = f"{self.channel}::" if with_channel and self.channel else ""
        return f"{channel}{self.project_name}{self.specifier}{self._format_marker()}"

    def _hash_key(self) -> tuple:
        return (
            self.key,
            frozenset(self.specifier),
        )

    def as_named_requirement(self) -> NamedRequirement:
        return NamedRequirement.create(name=self.name, marker=self.marker, specifier=self.specifier)


def wrap_from_req_dict(func):
    @functools.wraps(func)
    def wrapper(name: str, req_dict: RequirementDict) -> Requirement:
        if isinstance(req_dict, dict) and req_dict.get("conda_managed", False):
            return CondaRequirement.create(name=name, **req_dict)
        return func(name, req_dict)

    return wrapper


def parse_requirement(line: str, editable: bool = False) -> Requirement:
    m = _conda_req_re.match(line)
    if m is not None:
        if m is not None:
            line = m.group(1)
        channel = None
        if "::" in line:
            channel, line = line.split("::", maxsplit=1)
        line = re.sub(r"(>=?[\w.]+)\*", r"\g<1>0", line)
        if " " in line:
            name, version = line.split(" ", maxsplit=1)
            marker = ""
            if ";" in version:
                version, marker = line.split(";", maxsplit=1)
                marker = f";{marker}"
            for digit, s in re.findall(r"(\d)([a-z]+)", version):
                if s not in ("a", "b", "rc", "dev", "post"):
                    version = version.replace(f"{digit}{s}", digit)

            version = version.replace(",", " ")
            if not _specifier_re.match(version):
                version = f"=={version}"
            version = ",".join(v.strip() for v in version.split(" ") if _specifier_re.match(v))
            line = f"{name} {version} {marker}"
        prefix = ""
        if line.startswith("_"):
            prefix = "_"
            line = line[1:]

        package_req = PackageRequirement(line)  # type: ignore
        kwargs = dict(
            name=prefix + package_req.name,
            specifier=package_req.specifier,
            marker=get_marker(package_req.marker),
            channel=channel,
        )
        return CondaRequirement.create(**kwargs)

    return _parse_requirement(line=line, editable=editable)


if not _patched:
    from pdm.cli import actions
    from pdm.models import requirements

    setattr(Requirement, "from_req_dict", wrap_from_req_dict(Requirement.from_req_dict))
    setattr(actions, "parse_requirement", parse_requirement)
    setattr(requirements, "parse_requirement", parse_requirement)
    _patched = True



import argparse

from pdm.cli.commands.add import Command as BaseCommand
from pdm.cli.options import ArgumentGroup
from pdm.project import Project

conda_group = ArgumentGroup("Conda Arguments")
conda_group.add_argument(
    "--conda",
    dest="conda_packages",
    action="append",
    help="Specify Conda packages",
    default=[],
)

conda_group.add_argument(
    "-c",
    "--channel",
    dest="conda_channel",
    type=str,
    help="Specify Conda channel",
    default="",
)

conda_group.add_argument(
    "-r",
    "--runner",
    dest="conda_runner",
    type=str,
    help="Specify Conda runner executable",
    default="",
)


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "add"
    arguments = BaseCommand.arguments + [conda_group]

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        channel = options.conda_channel
        confs = project.pyproject.settings.setdefault("conda", dict())
        if options.conda_runner:
            confs["runner"] = options.conda_runner
        existing_channels = confs.setdefault("channels", [])
        conda_packages = options.conda_packages

        for package in conda_packages:
            if package.startswith("conda:"):
                package = package[len("conda:") :]
            package_channel = None
            if "::" not in package:
                if channel:
                    package_channel = channel
                    package = f"{channel}::{package}"
            else:
                package_channel, _ = package.split("::", maxsplit=1)
            if package_channel and package_channel not in existing_channels:
                project.core.ui.echo(f"Detected Conda channel {package_channel}, adding it to pyproject")
                existing_channels.append(package_channel)
            package = f"conda:{package}"
            options.packages.append(package)

        super().handle(project, options)



"""Configuration for the pytest test suite."""
from tempfile import TemporaryDirectory

import pytest
from pdm.cli.actions import do_init
from pdm.core import Core
from pdm.project import Project

from pdm_conda import main


@pytest.fixture(scope="session")
def core() -> Core:
    _core = Core()
    main(_core)
    yield _core


@pytest.fixture
def distributions(mocker):
    mocker.patch("pdm.models.working_set.distributions", return_value=[])


@pytest.fixture
def project(core, distributions) -> Project:
    with TemporaryDirectory() as tmp_dir:
        _project = core.create_project(tmp_dir)
        do_init(
            _project,
            name="test",
            version="0.0.0",
            python_requires=">=3.10",
            author="test",
            email="test@test.com",
        )
        yield _project


@pytest.fixture
def mock_conda(mocker, conda_response: dict | list, empty_conda_list):
    if isinstance(conda_response, dict):
        conda_response = [conda_response]
    install_response = {
        "actions": {
            "LINK": conda_response,
        },
    }

    def _mock(cmd, **kwargs):
        subcommand = cmd[1]
        if subcommand == "install":
            return install_response
        elif subcommand == "list":
            if empty_conda_list:
                return []
            return conda_response
        elif subcommand == "info":
            return {
                "virtual packages": [
                    "__unix=0=0",
                    "__linux=5.10.109=0",
                    "__glibc=2.35=0",
                    "__archspec=1=aarch64",
                ],
            }
        elif subcommand == "search":
            name = next(filter(lambda x: not x.startswith("-"), cmd[2:]))
            name = name.split(">")[0].split("<")[0].split("=")[0].split("~")[0]
            return {"result": {"pkgs": [p for p in conda_response if p["name"] == name]}}
        else:
            return {"message": "ok"}

    yield mocker.patch("pdm_conda.plugin.run_conda", side_effect=_mock)


PYTHON_REQUIREMENTS = [
    {
        "name": "lib2",
        "depends": [],
        "version": "1.0.0",
        "url": "https://channel.com/lib2",
        "channel": "https://channel.com",
        "sha256": "this-is-a-hash",
        "build_string": "lib2",
    },
    {
        "name": "lib",
        "depends": ["lib2 ==1.0.0"],
        "version": "1.0.0",
        "url": "https://channel.com/lib",
        "channel": "https://channel.com",
        "sha256": "this-is-a-hash",
        "build_string": "lib",
    },
    {
        "name": "python",
        "depends": ["lib ==1.0.0"],
        "version": "3.10.9",
        "url": "https://channel.com/python",
        "channel": "https://channel.com",
        "sha256": "this-is-a-hash",
        "build_string": "python",
    },
]

CONDA_INFO = [
    [
        *PYTHON_REQUIREMENTS,
        {
            "name": "another-dep",
            "depends": [],
            "version": "1.0.0",
            "url": "https://channel.com/another-dep",
            "channel": "https://channel.com",
            "sha256": "this-is-a-hash",
            "build_string": "another-dep",
        },
        {
            "name": "dep",
            "depends": ["python >=3.7", "another-dep ==1.0.0"],
            "version": "1.0.0",
            "url": "https://channel.com/dep",
            "channel": "https://channel.com",
            "sha256": "this-is-a-hash",
            "build_string": "dep",
        },
    ],
]



from typing import Any

from pdm.models.setup import Setup, SetupDistribution


class CondaSetupDistribution(SetupDistribution):
    def __init__(self, data: Setup, extras: dict | None = None) -> None:
        self.extras = extras or dict()
        super().__init__(data)

    @property
    def metadata(self) -> dict[str, Any]:
        metadata = super().metadata
        metadata["Home-Page"] = None
        metadata["License"] = "other"
        return metadata



from typing import cast

import pytest

from tests.conftest import CONDA_INFO, PYTHON_REQUIREMENTS


class TestAddRemove:
    conda_runner = "micromamba"

    @pytest.mark.parametrize("conda_response", CONDA_INFO)
    @pytest.mark.parametrize("empty_conda_list", [False])
    @pytest.mark.parametrize("packages", [["dep"], ["dep", "another-dep"], ["channel::dep", "another-dep"]])
    @pytest.mark.parametrize("channel", [None, "another_channel"])
    @pytest.mark.parametrize("runner", [None, "micromamba"])
    def test_add(self, core, project, mock_conda, conda_response, packages, channel, runner):
        """
        Test `add` command work as expected
        """
        from pdm_conda.models.config import PluginConfig
        from pdm_conda.project import CondaProject

        project.pyproject._data.update(
            {
                "tool": {
                    "pdm": {
                        "conda": {
                            "runner": runner or self.conda_runner,
                        },
                    },
                },
            },
        )
        command = ["add", "-v", "--no-self", "--no-sync"]
        for package in packages:
            command.extend(["--conda", package])
        if channel:
            command += ["--channel", channel]
        if runner:
            command += ["--runner", runner]
        else:
            runner = self.conda_runner
        core.main(command, obj=project)

        project.pyproject.reload()
        channels = set(p.split("::")[0] for p in packages if "::" in p)
        if channel:
            channels.add(channel)

        conf = PluginConfig.load_config(project)

        assert channels.issubset(conf.channels)
        assert conf.runner == runner

        num_search = 1  # add conda info
        packages_names = {p.split("::")[-1] for p in packages}
        for c in conda_response:
            if c["name"] in packages_names:
                num_search += 1 + len(c["depends"])
        assert mock_conda.call_count == num_search

        project = cast(CondaProject, project)
        dependencies = project.get_conda_pyproject_dependencies("default")
        for package in packages:
            _package = package.split("::")[-1]
            assert any(True for d in dependencies if _package in d)

    @pytest.mark.parametrize("conda_response", CONDA_INFO)
    @pytest.mark.parametrize("empty_conda_list", [False])
    @pytest.mark.parametrize("packages", [["dep"], ["dep", "another-dep"], ["channel::dep"]])
    def test_remove(self, core, project, mock_conda, conda_response, packages):
        self.test_add(core, project, mock_conda, conda_response, packages, None, None)
        mock_conda.reset_mock()
        core.main(["remove", "--no-self"] + packages, obj=project)
        conda_calls = len(conda_response) - len(PYTHON_REQUIREMENTS)
        cmd_order = []
        if conda_calls:
            cmd_order = ["list"] + ["search"] * len(PYTHON_REQUIREMENTS) + ["remove"] * conda_calls
        assert mock_conda.call_count == len(cmd_order)
        packages = [p["name"] for p in conda_response]
        python_packages = [f"{p['name']}=={p['version']}" for p in PYTHON_REQUIREMENTS]
        for (cmd,), kwargs in mock_conda.call_args_list:
            assert cmd[0] == self.conda_runner
            cmd_subcommand = cmd[1]
            assert cmd_subcommand == cmd_order.pop(0)
            if cmd_subcommand in ("remove", "search"):
                name = next(filter(lambda x: not x.startswith("-"), cmd[2:]))
                if cmd_subcommand == "remove":
                    assert name in packages
                    assert "-f" not in cmd
                elif cmd_subcommand == "search":
                    assert name in python_packages

        assert not cmd_order



from typing import Collection

from pdm.installers import Synchronizer
from pdm.models.candidates import Candidate
from pdm.models.environment import Environment

from pdm_conda.plugin import conda_search
from pdm_conda.utils import normalize_name


def _update_dependencies(name, packages, project, dependencies):
    if name in dependencies or name not in packages:
        return
    dependencies.add(name)
    dist = packages[name]
    candidates = conda_search(f"{dist.name}=={dist.version}", project, channel=dist.extras.get("channel", None))
    build_string = dist.extras["build_string"]
    candidate = next(c for c in candidates if c.build_string == build_string)
    for dep in candidate.dependencies:
        normalized_name = normalize_name(dep.name)
        _update_dependencies(normalized_name, packages, project, dependencies)
        dependencies.add(normalized_name)


class CondaSynchronizer(Synchronizer):
    def __init__(
        self,
        candidates: dict[str, Candidate],
        environment: Environment,
        clean: bool = False,
        dry_run: bool = False,
        retry_times: int = 1,
        install_self: bool = False,
        no_editable: bool | Collection[str] = False,
        use_install_cache: bool = False,
        reinstall: bool = False,
        only_keep: bool = False,
    ) -> None:
        super().__init__(
            candidates,
            environment,
            clean,
            dry_run,
            retry_times,
            install_self,
            no_editable,
            use_install_cache,
            reinstall,
            only_keep,
        )
        self.parallel = bool(self.parallel)  # type: ignore

    def compare_with_working_set(self) -> tuple[list[str], list[str], list[str]]:
        to_add, to_update, to_remove = super().compare_with_working_set()

        # get python dependencies and avoid removing them
        python_dependencies: set[str] = set()
        _update_dependencies("python", self.working_set, self.environment.project, python_dependencies)
        to_remove = [p for p in to_remove if p not in python_dependencies]
        # deactivate parallel execution if uninstall
        if to_remove or to_update:
            if self.parallel:
                self.environment.project.core.ui.echo("Deactivating parallel uninstall.")
            self.parallel = False
        else:
            self.parallel = self.environment.project.config["install.parallel"]

        return to_add, to_update, to_remove



import re
from importlib.metadata import Distribution
from pathlib import Path
from typing import Any, cast

from pdm.models.candidates import Candidate, PreparedCandidate
from pdm.models.environment import Environment
from pdm.models.setup import Setup
from unearth import Link

from pdm_conda.models.requirements import (
    CondaRequirement,
    Requirement,
    parse_requirement,
)
from pdm_conda.models.setup import CondaSetupDistribution


class CondaPreparedCandidate(PreparedCandidate):
    def __init__(self, candidate: Candidate, environment: Environment) -> None:
        super().__init__(candidate, environment)
        self.candidate = cast(CondaCandidate, self.candidate)  # type: ignore
        self.req = cast(CondaRequirement, self.req)  # type: ignore

    def get_dependencies_from_metadata(self) -> list[str]:
        # if conda candidate return already obtained dependencies
        return [d.as_line() for d in self.candidate.dependencies]

    def prepare_metadata(self) -> Distribution:
        # if conda candidate get setup from package
        return self.candidate.distribution


class CondaCandidate(Candidate):
    def __init__(
        self,
        req: Requirement,
        name: str | None = None,
        version: str | None = None,
        link: Link | None = None,
        dependencies: list[str] | None = None,
        build_string: str | None = None,
    ):
        super().__init__(req, name, version, link)
        # extract hash from link
        if link and link.hash is not None:
            self.hashes = {link: link.hash}
        self._req = cast(CondaRequirement, req)  # type: ignore
        self._preferred = None
        self._prepared: CondaPreparedCandidate | None = None
        self.dependencies = [parse_requirement(f"conda:{r}") for r in (dependencies or [])]
        self.build_string = build_string

    @property
    def req(self):
        return self._req

    @req.setter
    def req(self, value):
        if isinstance(value, CondaRequirement):
            self._req = value

    @property
    def distribution(self):
        return CondaSetupDistribution(
            Setup(
                name=self.name,
                summary="",
                version=self.version,
                install_requires=[d.as_line() for d in self.dependencies],
                python_requires=self.requires_python,
            ),
        )

    def as_lockfile_entry(self, project_root: Path) -> dict[str, Any]:
        result = super().as_lockfile_entry(project_root)
        result["conda_managed"] = True
        if self.req.channel is not None:
            result["channel"] = self.req.channel
        if self.link is None:
            raise ValueError("Uninitialized conda requirement")
        result["url"] = self.link.url
        if self.link.comes_from is not None:
            result["channel_url"] = self.link.comes_from
        return result

    def prepare(self, environment: Environment) -> CondaPreparedCandidate:
        """Prepare the candidate for installation."""
        if self._prepared is None:
            self._prepared = CondaPreparedCandidate(self, environment)
        return self._prepared

    @classmethod
    def from_lock_package(cls, package: dict) -> "CondaCandidate":
        """
        Create conda candidate from lockfile package.
        :param package: lockfile package
        :return: conda candidate
        """
        requires_python = package.get("requires_python", "")
        dependencies = package.get("dependencies", [])
        if requires_python:
            dependencies.append(f"python {requires_python}")
        return CondaCandidate.from_conda_package(
            package
            | {
                "channel": package.get("channel_url", None),
                "depends": dependencies,
                "build_string": None,
            },
        )

    @classmethod
    def from_conda_package(cls, package: dict) -> "CondaCandidate":
        """
        Create conda candidate from conda package.
        :param package: conda package
        :return: conda candidate
        """
        dependencies: list = package["depends"] or []
        requires_python = None
        to_delete = []
        for d in dependencies:
            if d.startswith("__"):
                to_delete.append(d)
            elif match := re.match(r"python( .+|$)", d):
                to_delete.append(d)
                if requires_python is None:
                    requires_python = match.group(1).strip() or None
        for d in to_delete:
            dependencies.remove(d)
        hashes = {h: package[h] for h in ["sha256", "md5"] if h in package}
        url = package["url"]
        for k, v in hashes.items():
            url += f"#{k}={v}"
        name, version = package["name"], package["version"]
        return CondaCandidate(
            req=parse_requirement(f"conda:{name} {version}"),
            name=name,
            version=version,
            link=Link(url, comes_from=package["channel"], requires_python=requires_python, hashes=hashes),
            dependencies=dependencies,
            build_string=package["build_string"],
        )



import pytest
from pdm.exceptions import ProjectError


class TestPluginConfig:
    @pytest.mark.parametrize("channels", [[], ["conda-forge"]])
    @pytest.mark.parametrize("runner", ["micromamba", "mamba", "conda", None])
    @pytest.mark.parametrize("as_default_manager", [True, False, None])
    def test_get_configs(self, project, channels, runner, as_default_manager):
        """
        Test loading configs correctly
        """
        dependencies = ["pytest"]
        optional_dependencies = {"dev": ["pytest"]}

        conf = {
            "dependencies": dependencies,
            "optional-dependencies": optional_dependencies,
        }
        if channels:
            conf["channels"] = channels
        else:
            channels = ["defaults"]

        if runner:
            conf["runner"] = runner
        else:
            runner = "conda"
        if isinstance(as_default_manager, bool):
            conf["as_default_manager"] = as_default_manager
        else:
            as_default_manager = False

        project.pyproject._data.update(
            {
                "tool": {
                    "pdm": {
                        "conda": conf,
                    },
                },
            },
        )
        from pdm_conda.models.config import PluginConfig

        config = PluginConfig.load_config(project)
        assert config.dependencies == dependencies
        assert config.optional_dependencies == optional_dependencies
        assert config.runner == runner
        assert config.channels == channels
        assert config.as_default_manager == as_default_manager

    def test_incorrect_runner(self, project):
        """
        Test load config raises on incorrect runner
        """
        runner = "another runner"
        project.pyproject._data.update(
            {
                "tool": {
                    "pdm": {
                        "conda": {
                            "runner": runner,
                        },
                    },
                },
            },
        )

        from pdm_conda.models.config import PluginConfig

        with pytest.raises(ProjectError, match=f"Invalid Conda runner: {runner}"):
            PluginConfig.load_config(project)



from typing import Any, Iterable, Mapping

from pdm.models.repositories import BaseRepository, LockedRepository, PyPIRepository
from pdm.models.requirements import Requirement
from pdm.models.specifiers import PySpecSet

from pdm_conda.models.candidates import Candidate, CondaCandidate
from pdm_conda.models.requirements import CondaRequirement


class CondaRepository(BaseRepository):
    def get_dependencies(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
        if isinstance(candidate, CondaCandidate):
            return (
                candidate.dependencies,
                PySpecSet(candidate.requires_python),
                candidate.summary,
            )
        return super().get_dependencies(candidate)


class PyPICondaRepository(CondaRepository, PyPIRepository):
    def _find_candidates(self, requirement: Requirement) -> Iterable[Candidate]:
        if isinstance(requirement, CondaRequirement):
            from pdm_conda.plugin import conda_search

            return conda_search(requirement, self.environment.project)
        return super()._find_candidates(requirement)


class LockedCondaRepository(CondaRepository, LockedRepository):
    def _read_lockfile(self, lockfile: Mapping[str, Any]) -> None:
        packages = lockfile.get("package", [])
        conda_packages = [p for p in packages if p.get("conda_managed", False)]
        packages = [p for p in packages if not p.get("conda_managed", False)]
        super()._read_lockfile({"package": packages})

        for package in conda_packages:
            can = CondaCandidate.from_lock_package(package)
            can_id = self._identify_candidate(can)
            self.packages[can_id] = can
            self.candidate_info[can_id] = (
                [d.as_line() for d in can.dependencies],
                package.get("requires_python", ""),
                "",
            )

    def _identify_candidate(self, candidate: Candidate) -> tuple:
        if isinstance(candidate, CondaCandidate):
            return candidate.identify(), candidate.version, None, False
        return super()._identify_candidate(candidate)



import pytest

from tests.conftest import CONDA_INFO, PYTHON_REQUIREMENTS
from tests.utils import format_url


class TestInstall:
    conda_runner = "micromamba"

    @pytest.mark.parametrize("conda_response", CONDA_INFO)
    @pytest.mark.parametrize("empty_conda_list", [True])
    @pytest.mark.parametrize("dry_run", [True, False])
    def test_install(self, core, project, mock_conda, conda_response, dry_run):
        """
        Test `install` command work as expected
        """
        project.pyproject._data.update(
            {
                "tool": {
                    "pdm": {
                        "conda": {
                            "runner": self.conda_runner,
                            "dependencies": [conda_response[-1]["name"]],
                        },
                    },
                },
            },
        )
        command = ["install", "-v", "--no-self"]
        if dry_run:
            command.append("--dry-run")
        core.main(command, obj=project)

        conda_calls = len(conda_response) - len(PYTHON_REQUIREMENTS)
        cmd_order = ["info"] + ["search"] * conda_calls + ["list"] + ["install"] * (0 if dry_run else conda_calls)
        assert mock_conda.call_count == len(cmd_order)

        urls = [format_url(p) for p in conda_response if p not in PYTHON_REQUIREMENTS]
        for (cmd,), kwargs in mock_conda.call_args_list:
            assert cmd[0] == self.conda_runner
            cmd_subcommand = cmd[1]
            assert cmd_subcommand == cmd_order.pop(0)
            if cmd_subcommand == "install":
                # if not install_dep:
                #     assert set(kwargs["dependencies"]) == requirements
                #     if not dry_run:
                #         install_dep = True
                # else:
                deps = kwargs["dependencies"]
                assert len(deps) == 1
                urls.remove(deps[0])

        if not dry_run:
            assert not urls
        assert not cmd_order



from contextlib import redirect_stdout
from io import StringIO

import pytest

from tests.conftest import CONDA_INFO


class TestList:
    conda_runner = "micromamba"

    @pytest.mark.parametrize("conda_response", CONDA_INFO)
    @pytest.mark.parametrize("empty_conda_list", [False])
    def test_list(self, core, project, mock_conda, conda_response):
        """
        Test `list` command work as expected
        """

        project.pyproject._data.update(
            {
                "tool": {
                    "pdm": {
                        "conda": {
                            "runner": self.conda_runner,
                            "dependencies": ["dep"],
                        },
                    },
                },
            },
        )
        command = ["list"]
        with StringIO() as output:
            with redirect_stdout(output):
                core.main(command, obj=project)

            o = output.getvalue()
            for p in conda_response:
                assert p["name"] in o

        assert mock_conda.call_count == 1



import argparse
from typing import cast

from pdm.cli.commands.remove import Command as BaseCommand
from pdm.project import Project

from pdm_conda.models.requirements import parse_requirement


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "remove"

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        from pdm_conda.project import CondaProject

        if options.group is None:
            options.group = "dev" if options.dev else "default"

        project = cast(CondaProject, project)
        conda_dependencies = project.get_conda_pyproject_dependencies(options.group, options.dev)
        dependencies, _ = project.get_pyproject_dependencies(options.group, options.dev)
        # add conda dependencies to common dependencies if going to remove it
        for i in range(len(options.packages)):
            # parse it as conda, if found add it to dependencies
            conda_package = f"conda:{options.packages[i]}"
            package = parse_requirement(conda_package)
            idx = None
            for dep_idx, dep in enumerate(conda_dependencies):
                if package.name in dep:
                    if package.name not in dependencies:
                        options.packages[i] = conda_package
                    else:
                        options.packages.append(conda_package)
                    dependencies.append(f"conda:{dep}")
                    idx = dep_idx
                    break
            if idx is not None:
                conda_dependencies.pop(idx)
        super().handle(project, options)



import pytest

from tests.conftest import CONDA_INFO, PYTHON_REQUIREMENTS


class TestLock:
    conda_runner = "micromamba"

    @pytest.mark.parametrize("conda_response", CONDA_INFO)
    @pytest.mark.parametrize("empty_conda_list", [False])
    @pytest.mark.parametrize("group", ["default", "dev", None])
    @pytest.mark.parametrize("add_conflict", [True, False])
    def test_lock(self, core, project, mock_conda, conda_response, group, add_conflict, refresh=False):
        """
        Test lock command work as expected
        """
        from pdm_conda.models.requirements import CondaRequirement

        python_requirements = {c["name"] for c in PYTHON_REQUIREMENTS}
        conda_response = [c for c in conda_response if c["name"] not in python_requirements]
        package = conda_response[1]["name"]
        project.pyproject._data.update(
            {
                "tool": {
                    "pdm": {
                        "conda": {
                            "runner": self.conda_runner,
                            "dependencies": [package],
                        },
                    },
                },
            },
        )
        if add_conflict:
            project.pyproject._data.update(
                {
                    "project": {"dependencies": [conda_response[0]["name"]], "requires-python": ">=3.10"},
                },
            )

        requirements = [r.as_line() for r in project.get_dependencies().values() if isinstance(r, CondaRequirement)]
        cmd = ["lock", "-v"]
        if refresh:
            cmd.append("--refresh")
        core.main(cmd, obj=project)

        cmd_order = []
        packages_to_search = set(requirements)
        # not all packages conda managed so run pre_lock
        if add_conflict:
            cmd_order = ["create", "install", "remove"]
            packages_to_search.add(conda_response[0]["name"])
        cmd_order.append("info")
        for c in conda_response:
            for d in c["depends"]:
                packages_to_search.add(d.replace(" ", ""))
        cmd_order.extend(["search"] * len(packages_to_search))

        assert mock_conda.call_count == len(cmd_order)

        requirements.append(f"python=={project.python.version}")
        for (cmd,), kwargs in mock_conda.call_args_list:
            assert cmd[0] == self.conda_runner
            cmd_subcommand = cmd[1]
            assert cmd_subcommand == cmd_order.pop(0)
            if cmd_subcommand == "install":
                assert set(kwargs["dependencies"]) == set(requirements)
            elif cmd_subcommand == "search":
                name = next(filter(lambda x: not x.startswith("-"), cmd[2:]))
                assert name in packages_to_search

        packages_to_search = {p.split("=")[0] for p in packages_to_search}
        assert not cmd_order
        lockfile = project.lockfile
        packages = lockfile["package"]
        for p in packages:
            name = p["name"]
            assert name in packages_to_search

    @pytest.mark.parametrize("conda_response", CONDA_INFO)
    @pytest.mark.parametrize("empty_conda_list", [False])
    @pytest.mark.parametrize("group", ["default", "dev", None])
    def test_lock_refresh(self, core, project, mock_conda, conda_response, group):
        self.test_lock(core, project, mock_conda, conda_response, group, False)
        self.test_lock(core, project, mock_conda, conda_response, group, False, refresh=True)



from dataclasses import dataclass, field

from pdm.exceptions import ProjectError
from pdm.project import ConfigItem, Project


@dataclass
class PluginConfig:
    channels: list[str] = field(default_factory=list)
    runner: str = "conda"
    as_default_manager: bool = False
    dependencies: list[str] = field(default_factory=list, repr=False)
    optional_dependencies: dict[str, list] = field(default_factory=dict)
    dev_dependencies: dict[str, list] = field(default_factory=dict)
    _initialized: bool = field(repr=False, default=False)

    def __post_init__(self):
        if not self.channels:
            self.channels = ["defaults"]
        if self.runner not in ["conda", "micromamba", "mamba"]:
            raise ProjectError(f"Invalid Conda runner: {self.runner}")

    @property
    def is_initialized(self):
        return self._initialized

    @classmethod
    def load_config(cls, project: Project, **kwargs) -> "PluginConfig":
        """
        Load plugin configs from project settings.
        :param project: Project
        :param kwargs: settings overwrites
        :return: plugin configs
        """
        config = {k.replace("-", "_"): v for k, v in project.pyproject.settings.get("conda", {}).items()}
        kwargs["_initialized"] = "conda" in project.pyproject.settings
        return PluginConfig(**(config | kwargs))

    def command(self, cmd="install"):
        """
        Get runner command args
        :param cmd: command, install by default
        :return: args list
        """
        _command = [self.runner, cmd, "-y"]
        if cmd in ("install", "create", "search"):
            _command.append("--strict-channel-priority")
        return _command

    @classmethod
    def configs(cls):
        _configs = [
            ("runner", ConfigItem("Conda runner executable", "conda")),
            ("channels", ConfigItem("Conda channels to use", ["defaults"])),
            ("as_default_manager", ConfigItem("Use Conda to install all possible requirements", False)),
            ("dependencies", ConfigItem("Dependencies to install with Conda", [])),
            ("optional-dependencies", ConfigItem("Optional dependencies to install with Conda", [])),
            ("dev-dependencies", ConfigItem("Development dependencies to install with Conda", [])),
        ]
        return [(f"conda.{name}", config) for name, config in _configs]



import re

import pytest

DEPENDENCIES = dict(
    argnames=["dependencies", "conda_dependencies"],
    argvalues=[
        (["pytest"], ["pytest-cov"]),
        (["pytest>=3.1"], ["pytest"]),
        (["pytest[extra]"], ["pytest"]),
        (["pytest"], ["pytest[extra]"]),
        (["pytest>=3.1"], ["pytest==52"]),
        ([], ["conda-channel::pytest"]),
        (["pytest"], []),
        ([], ["pytest>=1.*"]),
    ],
    ids=[
        "different pkgs",
        "pypi version override",
        "conda extras override",
        "conda no extras",
        "conda version override",
        "conda channel",
        "only pypi",
        "star greater specifier",
    ],
)
GROUPS = dict(argnames="group", argvalues=["default", "dev", "optional"])


class TestProject:
    def _parse_requirements(self, dependencies, conda_dependencies, as_default_manager=False):
        from pdm_conda.models.requirements import CondaRequirement, parse_requirement

        requirements = dict()
        for d in dependencies:
            if as_default_manager:
                if "[" in d:
                    d = d.split("[")[0]
                d = f"conda:{d}"
            r = parse_requirement(d)
            r.extras = None
            requirements[r.identify()] = r
        for d in conda_dependencies:
            r = parse_requirement(f"conda:{d}")
            if "::" in d:
                assert d.endswith(r.as_line())
            elif "[" not in d:
                assert r.as_line() == re.sub(r"(>=?[\w.]+)\*", r"\g<1>0", d)
            else:
                assert d.startswith(r.as_line())
            assert isinstance(r, CondaRequirement)
            assert not r.extras
            if "::" in d:
                assert r.channel == d.split("::")[0]
            pypi_req = next((v for v in requirements.values() if v.name == r.name), None)
            if pypi_req is not None:
                requirements.pop(pypi_req.identify())
                if not r.specifier:
                    r.specifier = pypi_req.specifier
            requirements[r.identify()] = r
        return requirements

    @pytest.mark.parametrize(**DEPENDENCIES)
    @pytest.mark.parametrize(**GROUPS)
    @pytest.mark.parametrize("as_default_manager", [False, True], ids=["", "as_default_manager"])
    def test_get_dependencies(self, project, dependencies, conda_dependencies, group, as_default_manager):
        """
        Test get project dependencies with conda dependencies and correct parse requirements
        """

        def dependencies_conf(dependencies, group):
            if group == "default":
                return {"dependencies": dependencies}
            else:
                return {f"{group}-dependencies": {"dev": dependencies}}

        def project_conf(dependencies, conda_dependencies, group, as_default_manager):
            dependencies, conda_dependencies = dependencies_conf(dependencies, group), dependencies_conf(
                conda_dependencies,
                group,
            )

            if group == "dev":
                dependencies.update({"conda": conda_dependencies})
                conf = {"tool": {"pdm": dependencies}}
            else:
                conf = {
                    "project": dependencies,
                    "tool": {
                        "pdm": {"conda": conda_dependencies},
                    },
                }
            if as_default_manager:
                conf["tool"]["pdm"].setdefault("conda", dict())["as_default_manager"] = True

            return conf

        project.pyproject._data.update(project_conf(dependencies, conda_dependencies, group, as_default_manager))

        if group != "default":
            group = "dev"

        requirements = self._parse_requirements(dependencies, conda_dependencies, as_default_manager)
        project_requirements = project.get_dependencies(group)
        for name, req in project_requirements.items():
            assert req == requirements[name]
            assert isinstance(req, type(requirements[name]))
        assert all("[" not in k for k in project_requirements)

    @pytest.mark.parametrize(**DEPENDENCIES)
    @pytest.mark.parametrize(**GROUPS)
    @pytest.mark.parametrize("python_packages", [True, False])
    def test_add_dependencies(self, project, dependencies, conda_dependencies, group, python_packages):
        requirements = self._parse_requirements(dependencies, conda_dependencies)

        group_name = group if group == "default" else "dev"
        dev = group == "dev"
        project.add_dependencies(requirements, to_group=group_name, dev=dev)
        project_requirements = project.get_dependencies(group_name)
        for name, req in project_requirements.items():
            assert req == requirements[name]
            assert isinstance(req, type(requirements[name]))
        if conda_dependencies and python_packages:
            asserted = 0
            _dependencies, _ = project.get_pyproject_dependencies(group_name, dev)
            _conda_dependencies = project.get_conda_pyproject_dependencies(group_name, dev)
            for d in conda_dependencies:
                d = d.split("[")[0].split("=")[0]
                for r in _dependencies:
                    if d.split("::")[-1] in r:
                        asserted += 1
                        break
                for r in _conda_dependencies:
                    if d in r:
                        asserted += 1
                        break
            assert asserted == len(conda_dependencies) * 2
        assert all("[" not in k for k in project.get_dependencies(group_name))



import functools

from pdm.models.environment import Environment

from pdm_conda.plugin import conda_list

_patched = False


def wrap_get_working_set(func):
    @functools.wraps(func)
    def wrapper(self: Environment):
        working_set = func(self)
        working_set._dist_map.update(conda_list(self.project))
        return working_set

    return wrapper


if not _patched:
    setattr(Environment, "get_working_set", wrap_get_working_set(Environment.get_working_set))
    _patched = True
